/* Copyright (C) 2016 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */
package sjc.backend.dennisk;

import sjc.backend.Architecture;
import sjc.backend.Instruction;
import sjc.compbase.Context;
import sjc.compbase.Marks;
import sjc.compbase.Mthd;
import sjc.compbase.Ops;
import sjc.compbase.StdTypes;
import sjc.osio.TextBuffer;
import sjc.osio.TextPrinter;

/**
 * MyVCPU2: Architecture backend for vCPU2 on MyCPU by Dennis Kuschel
 * 
 * @author S. Frenz
 * @version 160328 some more working
 *  version 160224 initial version
 */
public class MyVCPU2 extends Architecture {
  //basic concept:
  //  register d0 to d3 are used as 32 bit registers (id: 0..3, R00-07)
  //           p0 contains the current class         (id:    4, R08+09)
  //           p1 contains the current instance      (id:    5, R10+11)
  //           p2 is a temporary pointer             (id:    6, R12+13)
  //           sp contains the stack pointer         (id:    7, R14+15)
  
  //special "register"
  public final static int RegBaseMark = 0xAFFECAFE; //this mask should never occur
  
  //definitions of memory mapped registers
  public final static int RegMax   = 6;
  public final static int RegClss  = 4;
  public final static int RegInst  = 5;
  public final static int RegTmp   = 6;
  public final static int RegMaxGP = 3;
  
  public final static int R_D0_LO    =  0;
  public final static int R_D0_HI    =  1;
  public final static int R_D1_LO    =  2;
  public final static int R_D1_HI    =  3;
  public final static int R_D2_LO    =  4;
  public final static int R_D2_HI    =  5;
  public final static int R_D3_LO    =  6;
  public final static int R_D3_HI    =  7;
  public final static int R_CLSS_LO  =  8;
  public final static int R_CLSS_HI  =  9;
  public final static int R_INST_LO  = 10;
  public final static int R_INST_HI  = 11;
  public final static int R_TMP_LO   = 12;
  public final static int R_TMP_HI   = 13;
  public final static int R_STACK_LO = 14;
  public final static int R_STACK_HI = 15;
  
  public final static int R_D0_32    = R_D0_LO;
  public final static int R_D1_32    = R_D1_LO;
  public final static int R_D2_32    = R_D2_LO;
  public final static int R_D3_32    = R_D3_LO;
  public final static int R_CLSS_32  = R_CLSS_LO;
  public final static int R_INST_32  = R_INST_LO;
  public final static int R_TMP_32   = R_TMP_LO;
  public final static int R_STACK_32 = R_STACK_LO;
  
  public final static int RegMaskCLSS = (1<<RegClss);
  public final static int RegMaskINST = (1<<RegInst);
  public final static int RegMaskGP   = (1<<(RegMaxGP+1))-1;
  
  public final static int PAR_P1M     = 0x000F0000;
  public final static int PAR_REG1    = 0x00010000;
  public final static int PAR_IMM1    = 0x00020000;
  public final static int PAR_ADR1    = 0x00030000;
  public final static int PAR_P2M     = 0x00F00000;
  public final static int PAR_REG2    = 0x00100000;
  public final static int PAR_IMM2    = 0x00200000;
  public final static int PAR_ADR2    = 0x00300000;
  public final static int SIZE_MASK   = 0x0F000000;
  public final static int SIZE_SHIFT  = 24;
  public final static int SIZE_1      = 0x01000000;
  public final static int SIZE_2      = 0x02000000;
  public final static int SIZE_4      = 0x04000000;
  
  public final static int IT_CALL = -1;
  
  public final static int I_RET      = 0x01 | SIZE_2;                       //return from subroutine
  public final static int I_JPNZra   = 0x14 | PAR_REG1 | PAR_ADR2 | SIZE_4; //jump to address if register is not zero
  public final static int I_DECri    = 0x3D | PAR_REG1 | PAR_IMM2 | SIZE_2; //decrease register by given immediate (dec: 1..15)
  public final static int I_INCDri   = 0x3E | PAR_REG1 | PAR_IMM2 | SIZE_2; //increase 32-bit register by given immediate (inc: 1..15) 
  public final static int I_PUSHrr   = 0x43 | PAR_REG1 | PAR_REG2 | SIZE_2; //push content of register range
  public final static int I_POPrr    = 0x44 | PAR_REG1 | PAR_REG2 | SIZE_2; //pop content of register range
  public final static int I_XORrr    = 0x52 | PAR_REG1 | PAR_REG2 | SIZE_2; //logical xor of two registers
  public final static int I_ADDDrr   = 0x4D | PAR_REG1 | PAR_REG2 | SIZE_2; //add two 32 bit registers without carry
  public final static int I_LDri     = 0x77 | PAR_REG1 | PAR_IMM2 | SIZE_4; //load immediate value to register
  
  public final static int VAROFF_PARAM_NRM = 3;
  public final static int VAROFF_PARAM_INL = 1;
  
  public final static int STACKMEM_OFF = 0x0101;
  
  protected final static String ERR_INVGLOBADDRREG = "invalid register for address";
  
  private Mthd mthdContainer;
  private int usedRegs, writtenRegs, nextAllocReg, curVarOffParam;
  private TextBuffer asmTmpTextBuffer;
  
  //initialization
  public MyVCPU2() {
    relocBytes=4;
    stackClearBits=1;
    allocClearBits=3;
    maxInstrCodeSize=4;
    throwFrameSize=12;
    throwFrameExcOff=10;
    regClss=RegMaskCLSS;
    regInst=RegMaskINST;
    regBase=RegBaseMark;
    binAriCall[StdTypes.T_BYTE]|=(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_SHRT]|=(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_INT]|=(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_LONG]|=(1<<Ops.A_MUL-Ops.MSKBSE)|(1<<(Ops.A_DIV-Ops.MSKBSE))|(1<<(Ops.A_MOD-Ops.MSKBSE));
    binAriCall[StdTypes.T_FLT]|=(1<<(Ops.A_MUL-Ops.MSKBSE))|(1<<(Ops.A_DIV-Ops.MSKBSE))
        |(1<<(Ops.A_MOD-Ops.MSKBSE))|(1<<(Ops.A_PLUS-Ops.MSKBSE))|(1<<(Ops.A_MINUS-Ops.MSKBSE));
    binAriCall[StdTypes.T_DBL]|=(1<<(Ops.A_MUL-Ops.MSKBSE))|(1<<(Ops.A_DIV-Ops.MSKBSE))
        |(1<<(Ops.A_MOD-Ops.MSKBSE))|(1<<(Ops.A_PLUS-Ops.MSKBSE))|(1<<(Ops.A_MINUS-Ops.MSKBSE));
    unaAriCall[StdTypes.T_FLT]|=(1<<(Ops.A_MINUS-Ops.MSKBSE));
    unaAriCall[StdTypes.T_DBL]|=(1<<(Ops.A_MINUS-Ops.MSKBSE));
  }
  
  public String checkBuildAssembler(Context preInitCtx) {
    asmTmpTextBuffer=new TextBuffer();
    return "vCPU2";
  }
  
  protected void attachMethodAssemblerText(Mthd generatingMthd, Instruction first) {
    printCode(asmTmpTextBuffer, first, null, true);
    generatingMthd.asmCode=asmTmpTextBuffer.toString();
    asmTmpTextBuffer.reset();
  }
  
  //general memory and register management
  public void putRef(Object loc, int offset, Object ptr, int ptrOff) {
    mem.putInt(loc, offset, mem.getAddrAsInt(ptr, ptrOff));
  }

  public void putCodeRef(Object loc, int offset, Object ptr, int ptrOff) {
    mem.putInt(loc, offset, mem.getAddrAsInt(ptr, ptrOff));
  }
  
  private int bitSearch(int value, int hit) {
    int i, j;
    for (i=0; i<=RegMax; i++) {
      j=1<<i;
      if ((value&j)!=0) {
        if (--hit==0) return j;
      }
    }
    //nothing found, sorry
    return 0;
  }
  
  private int getBitPos(int reg) {
    int i, j;
    for (i=0; i<32; i++) { //also search "special registers"
      j=1<<i;
      if ((reg&j)!=0) return i;
    }
    return -1;
  }
  
  private int freeRegSearch(int mask, int type) {
    int ret=0, temp, regCount;
    regCount=((type==StdTypes.T_LONG)||(type==StdTypes.T_DBL)) ? 2 : 1;
    if (regCount==1) return bitSearch(mask, 1);
    while (regCount-->0) {
      if ((temp=bitSearch(mask, 1))==0) return 0;
      ret|=temp;
      mask&=~temp;
    }
    return ret;
  }
  
  protected int storeReg(int regs) {
    regs&=usedRegs&writtenRegs;
    for (int i=0; i<=RegMaxGP; i++) if ((regs&(1<<i))!=0) {
      int start=i;
      while (++i<=RegMaxGP && (regs&(1<<i))!=0);
      ins(I_PUSHrr, start, i-1);
    }
    return regs;
  }
  
  protected void restoreReg(int regs) {
    usedRegs|=regs;
    writtenRegs|=regs;
    for (int i=RegMaxGP; i>=0; i--) if ((regs&(1<<i))!=0) {
      int end=i;
      while (--i>=0 && (regs&(1<<i))!=0);
      ins(I_POPrr, i+1, end);
    }
  }
  
  public int ensureFreeRegs(int ignoreReg1, int ignoreReg2, int keepReg1, int keepReg2) {
    int restore=storeReg(RegMaskGP&~(ignoreReg1|ignoreReg2));
    usedRegs=(keepReg1|keepReg2)&RegMaskGP;
    return restore;
  }
  
  public int prepareFreeReg(int avoidReg1, int avoidReg2, int reUseReg, int type) {
    int toStore, ret;
    //basereg can not be allocated
    if (avoidReg1==regBase) avoidReg1=0;
    if (avoidReg2==regBase) avoidReg2=0;
    if (reUseReg==regBase) reUseReg=0;
    //first: try to reuse given regs
    reUseReg&=RegMaskGP;
    if (reUseReg!=0) {
      if ((ret=freeRegSearch(reUseReg, type))!=0) {
        usedRegs|=(nextAllocReg=ret);
        return 0; //nothing has to be freed, reuse given registers
      }
    }
    //second: try to alloc everywhere normally
    if ((ret=freeRegSearch((RegMaskGP&~usedRegs&~(avoidReg1|avoidReg2))|reUseReg, type))!=0) {
      usedRegs|=(nextAllocReg=ret);
      return 0; //nothing has to be freed, use newly allocated registers
    }
    //third: try to free a register
    if ((ret=freeRegSearch((RegMaskGP&~(avoidReg1|avoidReg2))|reUseReg, type))!=0) {
      toStore=storeReg(ret);
      deallocReg(toStore);
      usedRegs|=(nextAllocReg=ret);
      return toStore;
    }
    //no possibility found to free registers
    fatalError("unsolvable register request");
    return 0;
  }
  
  protected void deallocReg(int regs) {
    usedRegs&=~regs;
    writtenRegs&=~regs;
  }
  
  public int allocReg() {
    return nextAllocReg;
  }
  
  public void deallocRestoreReg(int deallocRegs, int keepRegs, int restore) {
    deallocReg(deallocRegs&~(keepRegs|restore));
    restoreReg(restore);
    usedRegs|=(keepRegs|restore)&RegMaskGP;
  }
  
  //method start, end and finalization
  public Mthd prepareMethodCoding(Mthd mthd) {
    Mthd lastMthd;
    if ((lastMthd=curMthd)!=null) curInlineLevel++;
    else mthdContainer=mthd; //remember outest level method
    writtenRegs=usedRegs=0;
    if ((lastMthd=curMthd)!=null) {
      curVarOffParam=VAROFF_PARAM_INL;
      curInlineLevel++;
      curVarOffParam=0;
    }
    else curVarOffParam=VAROFF_PARAM_NRM;
    curMthd=mthd;
    return lastMthd;
  }

  public void codeProlog() {
    Instruction loopDest;
    int i;
    if (curInlineLevel>0 && curMthd.parSize==0 && curMthd.varSize==0) { //optimize inline methods without parameters and variables
      return;
    }
    if (curMthd.parSize+curMthd.varSize>253) {
      fatalError("method has more than 253 bytes var+par on stack");
      return;
    }
    if ((curMthd.marker&Marks.K_INTR)!=0) {
      if (curInlineLevel>0) {
        fatalError("interrupt method can not be inlined");
        return;
      }
      if (curMthd.parSize!=0) {
        fatalError("interrupt method can not have parameters");
        return;
      }
      fatalError("interrupt methods not supported yet in codeProlog");
    }
    if (curMthd.varSize>0) {
      ins(I_XORrr, R_TMP_LO, R_TMP_LO);
      if (curMthd.varSize<=8) for (i=0; i<curMthd.varSize; i+=2) ins(I_PUSHrr, R_TMP_LO, R_TMP_LO);
      else {
        ins(I_LDri, R_TMP_HI, curMthd.varSize/2);
        appendInstruction(loopDest=getUnlinkedInstruction());
        ins(I_PUSHrr, R_TMP_LO, R_TMP_LO);
        ins(I_DECri, R_TMP_HI, 1);
        insJmp(I_JPNZra, R_TMP_HI, loopDest);
      }
    }
  }

  public void codeEpilog(Mthd outline) {
    if (outline!=null && curMthd.parSize==0 && curMthd.varSize==0) { //optimize inline methods without parameters and variables
      --curInlineLevel;
      curMthd=outline;
      return;
    }
    if (curMthd.varSize!=0) {
      if (curMthd.varSize<=15) ins(I_INCDri, R_STACK_32, curMthd.varSize);
      else {
        ins(I_LDri, R_TMP_LO, curMthd.varSize&0xFFFF);
        ins(I_LDri, R_TMP_HI, curMthd.varSize>>>16);
        ins(I_ADDDrr, R_STACK_32, R_TMP_32);
      }
    }
    if ((curMthd.marker&Marks.K_INTR)!=0) {
      fatalError("interrupt methods not supported yet in codeEpilog");
      curMthd=null;
    }
    else if (outline!=null) {
      if (--curInlineLevel==0) curVarOffParam=VAROFF_PARAM_NRM;
      --curInlineLevel;
      curMthd=outline;
    }
    else {
      ins(I_RET);
      curMthd=null;
    }
  }
  
  public void finalizeInstructions(Instruction first) {
    //nothing to do, all instructions are ready to be copied, only jumps/magics need to be fixed after address is known
    //print disassembly if wanted
    if (ctx.printCode || (mthdContainer.marker&Marks.K_PRCD)!=0)
      printCode(ctx.out, first, "X86", false);
  }

  public void finalizeInstructionAddresses(Mthd generatingMthd, Instruction first, Object loc, int offset) {
    int baseInlineCodeAddress=-1;
    Instruction now=first;
    int base=mem.getAddrAsInt(loc, offset);
    if ((generatingMthd.marker&Marks.K_DEBG)!=0) {
      ctx.out.print(" ---> base code address of debugged method ");
      generatingMthd.printNamePar(ctx.out);
      ctx.out.print(" is 0x");
      ctx.out.printHexFix(base, 4);
      ctx.out.println();
    }
  }
  
  private int getRelativeAddressOfInstruction(Instruction first, Instruction dest) {
    Instruction now=first;
    int offset=0;
    while (now!=null) {
      if (now==dest) return offset;
      offset+=now.size;
      now=now.next;
    }
    fatalError("could not find relative address of instruction");
    return 0;
  }
  
  //basic architecture implementation
  public void genLoadConstVal(int dst, int val, int type) {
    int reg, count, i;
    if (dst==regBase) {
      fatalError("invalid use of BaseReg in genLoadVarAddr");
      return;
    }
    /*if ((count=getByteCount(type))==-1) return;
    for (i=0; i<count; i++) {
      if ((reg=getRegZPAddr(i+1, dst, type, true))==-1) return;
      ins(I_MOVzpimm, reg, (val>>>(i<<3))&0xFF);
    }*/
  }

  public void genLoadConstDoubleOrLongVal(int dst, long val, boolean asDouble) {
    int reg, i, tempVal;
    if (dst==regBase) {
      fatalError("invalid use of BaseReg in genLoadVarAddr");
      return;
    }
    /*for (i=0; i<8; i++) {
      if ((reg=getRegZPAddr(i+1, dst, StdTypes.T_LONG, true))==-1) return;
      tempVal=(int)(val>>>(i<<3))&0xFF; //determine value to load
      if (tempVal==0) ins(I_CLA);
      else ins(I_LDAimm, tempVal);
      ins(I_STAzp, reg);
    }*/
  }

  public void genLoadVarAddr(int dst, int src, Object loc, int off) {
    int pos=mem.getAddrAsInt(loc, off);
    int dst1, dst2;
    if (dst==regBase) {
      fatalError("invalid use of BaseReg in genLoadVarAddr");
      return;
    }
    if (pos==0 && src==dst) return;
    /*if ((dst1=getRegZPAddr(1, dst, StdTypes.T_PTR, true))==-1
        || (dst2=getRegZPAddr(2, dst, StdTypes.T_PTR, true))==-1) return;
    if (src==regBase) {
      if (pos>=0) pos+=curVarOffParam;
      pos+=STACKMEM_OFF;
      ins(I_TXA);
      ins(I_CLC);
      ins(I_ADCimm, pos&0xFF);
      ins(I_STAzp, dst1);
      ins(I_CLA);
      ins(I_ADCimm, (pos>>>8)&0xFF);
      ins(I_STAzp, dst2);
    }
    else if (src!=0) {
      int src1, src2;
      if ((src1=getRegZPAddr(1, src, StdTypes.T_PTR, true))==-1
          || (src2=getRegZPAddr(2, src, StdTypes.T_PTR, true))==-1) return;
      ins(I_LDAzp, src1);
      if (pos!=0) {
        ins(I_CLC);
        if ((pos&0xFF)!=0) ins(I_ADCimm, pos&0xFF);
      }
      ins(I_STAzp, dst1);
      ins(I_LDAzp, src2);
      if (pos!=0) ins(I_ADCimm, (pos>>>8)&0xFF);
      ins(I_STAzp, dst2);
    }
    else {
      ins(I_LDAimm, pos&0xFF);
      ins(I_STAzp, dst1);
      ins(I_LDAimm, (pos>>>8)&0xFF);
      ins(I_STAzp, dst2);
    }*/
  }

  public void genLoadVarVal(int dst, int src, Object loc, int off, int type) {
    int dstReg, count, i=0, pos=mem.getAddrAsInt(loc, off), src1, src2;;
    /*count=getByteCount(type);
    if (src==regBase) {
      if (pos>=0) pos+=curVarOffParam;
      pos+=STACKMEM_OFF;
      while (i<count) {
        ins(I_LDAabsx, pos+i);
        if ((dstReg=getRegZPAddr(++i, dst, type, true))==-1) return;
        ins(I_STAzp, dstReg);
      }
      return;
    }
    ins(I_PHX);
    if (src!=0) {
      if ((src1=getRegZPAddr(1, src, StdTypes.T_PTR, true))==-1
          || (src2=getRegZPAddr(2, src, StdTypes.T_PTR, true))==-1) return;
      if (pos!=0) {
        ins(I_CLC);
        ins(I_LDAzp, src1);
        ins(I_ADCimm, pos&0xFF);
        ins(I_TAX);
        ins(I_LDAzp, src2);
        ins(I_ADCimm, (pos>>>8)&0xFF);
        ins(I_TAY);
      }
      else {
        ins(I_LDXzp, src1);
        ins(I_LDYzp, src2);
      }
    }
    else {
      ins(I_LDXimm, pos&0xFF);
      ins(I_LDYimm, (pos>>>8)&0xFF);
    }
    while (i<count) {
      if ((dstReg=getRegZPAddr(++i, dst, type, true))==-1) return;
      ins(I_LPA);
      ins(I_STAzp, dstReg);
    }
    ins(I_PLX);*/
  }

  public void genConvertVal(int dst, int src, int toType, int fromType) {
    int countFrom, countTo, regDst, regSrc=0, i=1, temp;
    Instruction dest;
    /*countFrom=getByteCount(fromType);
    countTo=getByteCount(toType);
    temp=countFrom>=countTo ? countTo : countFrom; //determine minimum value
    for (; i<=temp; i++) { //take care of lower bytes, regSrc is highest valid register
      if ((regDst=getRegZPAddr(i, dst, toType, true))==-1 || 
        (regSrc=getRegZPAddr(i, src, fromType, false))==-1) return;
      if (regDst!=regSrc) ins(I_MOVzpzp, regDst, regSrc);
    }
    if (countTo>countFrom) { //conversion was an expansion
      ins(I_CLA);
      if (fromType!=StdTypes.T_CHAR) { //not char -> sign extension
        ins(I_BITzpimm, regSrc, 0x80);
        insJmp(I_JPZimm, dest=getUnlinkedInstruction());
        ins(I_DEC);
        appendInstruction(dest);
      }
      //else //char -> zero extension
      for (; i<=countTo; i++) {
        if ((regDst=getRegZPAddr(i, dst, toType, true))==-1) return;
        ins(I_STAzp, regDst);
      }
    }*/
  }

  public void genDup(int dst, int src, int type) {
    int srcReg, dstReg, regCount, i;
    if (dst==src) return;
    /*regCount=getByteCount(type);
    for (i=1; i<=regCount; i++) {
      if ((dstReg=getRegZPAddr(i, dst, type, true))==-1
          || (srcReg=getRegZPAddr(i, src, type, false))==-1) return;
      if (dstReg!=srcReg) {
        ins(I_MOVzpzp, dstReg, srcReg);
      }
    }*/
  }

  public void genPushConstVal(int val, int type) {
    int count, shortVal;
    /*count=getByteCount(type);
    if (count==1) ins(I_PUSHimm, val&0xFF);
    else {
      if ((count&1)!=0) {
        fatalError("odd regCount>=2 in genPushConstVal");
        return;
      }
      while (count>0) {
        count-=2;
        shortVal=(val>>>(count<<3))&0xFFFF;
        ins(I_PUSAimm, shortVal);
      }
    }*/
  }

  public void genPushConstDoubleOrLongVal(long val, boolean asDouble) {
    int count=8, shortVal;
    /*while (count>0) {
      count-=2;
      shortVal=((int)(val>>>(count<<3)))&0xFFFF;
      ins(I_PUSAimm, shortVal);
    }*/
  }

  public void genPush(int src, int type) {
    int count, reg;
    /*count=getByteCount(type);
    while (count>0) {
      if ((reg=getRegZPAddr(count--, src, type, false))==-1) return;
      ins(I_PUSHzp, reg);
    }*/
  }

  public void genPop(int dst, int type) {
    int count, reg, i=0;
    /*count=getByteCount(type);
    while (i<count) {
      if ((reg=getRegZPAddr(++i, dst, type, true))==-1) return;
      ins(I_POPzp, reg);
    }*/
  }

  public void genAssign(int dst, int src, int type) {
    int i, regCount, reg1, reg2, srcReg;
    /*regCount=getByteCount(type);
    if ((reg1=getRegZPAddr(1, dst, StdTypes.T_PTR, false))==-1
        || (reg2=getRegZPAddr(2, dst, StdTypes.T_PTR, false))==-1) return;
    ins(I_PHX);
    ins(I_LDXzp, reg1);
    ins(I_LDYzp, reg2);
    for (i=1; i<=regCount; i++) {
      if ((srcReg=getRegZPAddr(i, src, type, false))==-1) return;
      ins(I_LDAzp, srcReg);
      ins(I_SPA);
    }
    ins(I_PLX);*/
  }

  public void genBinOp(int dst, int src1, int src2, int op, int type) {
    int opType=op>>>16, opPar=op&0xFFFF;
    int dstR, srcR1, srcR2, i, count;
    Instruction end, redo;
    /*count=getByteCount(type);
    switch (type) {
      case StdTypes.T_BOOL:
        if (opType!=Ops.S_ARI || !(opPar==Ops.A_AND || opPar==Ops.A_OR || opPar==Ops.A_XOR)) {
          fatalError("unsupported operation for bool-genBinOp");
          return;
        }
        //has to do the following, too!
      case StdTypes.T_BYTE:
        if (opPar==Ops.A_MUL) { //special case: mul for byte is supported
          if ((dstR=getRegZPAddr(1, dst, StdTypes.T_BYTE, true))==-1
              || (srcR1=getRegZPAddr(1, src1, StdTypes.T_BYTE, false))==-1
              || (srcR2=getRegZPAddr(1, src2, StdTypes.T_BYTE, false))==-1) return;
          ins(I_PHX);
          ins(I_LDAzp, srcR1);
          ins(I_MULzp, srcR2);
          ins(I_STAzp, dstR);
          ins(I_PLX);
          return;
        }
        //has to do the following, too!
      case StdTypes.T_SHRT: case StdTypes.T_CHAR:
      case StdTypes.T_INT:
      case StdTypes.T_LONG:
        switch (opType) {
          case Ops.S_ARI:
            for (i=1; i<=count; i++) {
              if ((dstR=getRegZPAddr(i, dst, type, true))==-1
                  || (srcR1=getRegZPAddr(i, src1, type, false))==-1
                  || (srcR2=getRegZPAddr(i, src2, type, false))==-1) return;
              ins(I_LDAzp, srcR1);
              switch (opPar) {
                case Ops.A_AND:
                  ins(I_ANDzp, srcR2);
                  break;
                case Ops.A_OR:
                  ins(I_ORzp, srcR2);
                  break;
                case Ops.A_XOR:
                  ins(I_EORzp, srcR2);
                  break;
                case Ops.A_MINUS:
                  if (i==1) ins(I_SEC);
                  ins(I_SBCzp, srcR2);
                  break;
                case Ops.A_PLUS:
                  if (i==1) ins(I_CLC);
                  ins(I_ADCzp, srcR2);
                  break;
                default:
                  fatalError("unsupported ari-operation for genBinOp");
                  return;
              }
              ins(I_STAzp, dstR);
            }
            return;
          case Ops.S_BSH:
            end=getUnlinkedInstruction();
            if ((srcR2=getRegZPAddr(1, src2, type, false))==-1) return;
            ins(I_LDYzp, srcR2);
            insJmp(I_JPZimm, end);
            redo=getUnlinkedInstruction();
            appendInstruction(redo);
            if ((dstR=getRegZPAddr(opPar==Ops.B_SHL ? 1 : count, dst, type, true))==-1
                || (srcR1=getRegZPAddr(opPar==Ops.B_SHL ? 1 : count, src1, type, false))==-1) return;
            switch (opPar) {
              case Ops.B_SHL:
                if (srcR1!=dstR) {
                  ins(I_LDAzp, srcR1);
                  ins(I_SHL);
                  ins(I_STAzp, dstR);
                }
                else ins(I_SHLzp, dstR);
                i=2;
                break;
              case Ops.B_SHRL:
                if (srcR1!=dstR) {
                  ins(I_LDAzp, srcR1);
                  ins(I_SHR);
                  ins(I_STAzp, dstR);
                }
                else ins(I_SHRzp, dstR);
                i=count-1;
                break;
              case Ops.B_SHRA:
                ins(I_LDAzp, srcR1);
                ins(I_SHL);
                ins(I_LDAzp, srcR1);
                ins(I_ROR);
                ins(I_STAzp, dstR);
                i=count-1;
                break;
              default:
                fatalError("unsupported bsh-operation for genBinOp");
                return;
            }
            while (--count>0) { //higher bytes - respect carry
              if ((dstR=getRegZPAddr(i, dst, type, true))==-1
                  || (srcR1=getRegZPAddr(i, src1, type, false))==-1) return;
              if (opPar==Ops.B_SHL) {
                if (dstR!=srcR1) {
                  ins(I_LDAzp, srcR1);
                  ins(I_ROL);
                  ins(I_STAzp, dstR);
                }
                else ins(I_ROLzp, dstR);
                i++;
              }
              else {
                if (dstR!=srcR1) {
                  ins(I_LDAzp, srcR1);
                  ins(I_ROR);
                  ins(I_STAzp, dstR);
                }
                else ins(I_RORzp, dstR);
                i--;
              }              
            }
            insJmp(I_DYJPimm, redo); //decrease counter and jump to redo
            appendInstruction(end);
            return;
          default:
            fatalError("unsupported operation-type for genBinOp");
            return;
        }
      default:
        fatalError("unsupported operand-type for genBinOp");
    }*/
  }

  public void genUnaOp(int dst, int src, int op, int type) {
    int opPar=op&0xFFFF, dstR, srcR, count, i=1;
    if (type==StdTypes.T_FLT || type==StdTypes.T_DBL) {
      fatalError("unary operator not yet supported for float and double");
      return;
    }
    /*if ((dstR=getRegZPAddr(1, dst, type, true))==-1
        || (srcR=getRegZPAddr(1, src, type, false))==-1) return;
    if (type==StdTypes.T_BOOL) {
      if (opPar==Ops.L_NOT) {
        ins(I_LDAzp, srcR);
        ins(I_EORimm, 1);
        ins(I_STAzp, dstR);
        return;
      }
      else {
        fatalError("unsupported bool-operation for genUnaOp");
        return;
      }
    }
    else if (opPar==Ops.A_MINUS) {
      ins(I_CLA);
      ins(I_SEC);
      ins(I_SBCzp, srcR);
      ins(I_STAzp, dstR);
      count=getByteCount(type);
      for (i=2; i<=count; i++) {
        if ((dstR=getRegZPAddr(i, dst, type, true))==-1
            || (srcR=getRegZPAddr(i, src, type, false))==-1) return;
        ins(I_CLA);
        ins(I_SBCzp, srcR);
        ins(I_STAzp, dstR);
      }
      return;
    }
    else if (opPar==Ops.A_CPL) {
      count=getByteCount(type);
      for (; i<=count; i++) {
        if ((dstR=getRegZPAddr(i, dst, type, true))==-1
            || (srcR=getRegZPAddr(i, src, type, false))==-1) return;
        ins(I_LDAzp, srcR);
        ins(I_EORimm, 0xFF);
        ins(I_STAzp, dstR);
      }
      return;
    }
    if (opPar!=Ops.A_PLUS) { //do nothing for Ops.A_PLUS
      fatalError("unsupported operation for genUnaOp");
    }*/
  }

  public void genIncMem(int dst, int type) {
    int i, regCount, reg1, reg2;
    /*regCount=getByteCount(type);
    if ((reg1=getRegZPAddr(1, dst, StdTypes.T_PTR, false))==-1
        || (reg2=getRegZPAddr(2, dst, StdTypes.T_PTR, false))==-1) return;
    ins(I_CLY);
    if (reg2!=reg1+1) {
      ins(I_MOVzpzp, RegTmpLo, reg1);
      ins(I_MOVzpzp, RegTmpHi, reg2);
      reg1=RegTmpLo;
      reg2=RegTmpHi;
    }
    if (regCount==1) {
      ins(I_INCizpy, reg1);
      return;
    }
    ins(I_CLC);
    ins(I_LDAizpy, reg1);
    ins(I_ADCimm, 1);
    ins(I_STAizpy, reg1);
    for (i=1; i<regCount; i++) {
      ins(I_INY);
      ins(I_LDAizpy, reg1);
      ins(I_ADCimm, 0);
      ins(I_STAizpy, reg1);
    }*/
  }

  public void genDecMem(int dst, int type) {
    /*int i, regCount, reg1, reg2;
    regCount=getByteCount(type);
    if ((reg1=getRegZPAddr(1, dst, StdTypes.T_PTR, false))==-1
        || (reg2=getRegZPAddr(2, dst, StdTypes.T_PTR, false))==-1) return;
    ins(I_CLY);
    if (reg2!=reg1+1) {
      ins(I_MOVzpzp, RegTmpLo, reg1);
      ins(I_MOVzpzp, RegTmpHi, reg2);
      reg1=RegTmpLo;
      reg2=RegTmpHi;
    }
    if (regCount==1) {
      ins(I_DECizpy, reg1);
      return;
    }
    ins(I_SEC);
    ins(I_LDAizpy, reg1);
    ins(I_SBCimm, 1);
    ins(I_STAizpy, reg1);
    for (i=1; i<regCount; i++) {
      ins(I_INY);
      ins(I_LDAizpy, reg1);
      ins(I_SBCimm, 0);
      ins(I_STAizpy, reg1);
    }*/
  }

  public void genSaveUnitContext() {
    ins(I_PUSHrr, R_CLSS_LO, R_CLSS_HI);
  }

  public void genRestUnitContext() {
    ins(I_POPrr, R_CLSS_LO, R_CLSS_HI);
  }

  public void genLoadUnitContext(int dst, int off) {
    fatalError("genLoadUnitContext not implemented yet");
  }

  public void genLoadConstUnitContext(int dst, Object unitLoc) {
    genLoadConstVal(dst, mem.getAddrAsInt(unitLoc, 0), StdTypes.T_SHRT);
  }

  public void genSaveInstContext() {
    ins(I_PUSHrr, R_INST_LO, R_INST_HI);
  }

  public void genRestInstContext() {
    ins(I_POPrr, R_INST_LO, R_INST_HI);
  }

  public void genLoadInstContext(int src) {
    /*int src1, src2;
    if ((src1=getRegZPAddr(1, src, StdTypes.T_PTR, false))==-1
        || (src2=getRegZPAddr(2, src, StdTypes.T_PTR, false))==-1) return;
    ins(I_MOVzpzp, ZPAddrInstLo, src1);
    ins(I_MOVzpzp, ZPAddrInstHi, src2);
    ins(I_PHX);
    ins(I_LDAzp, src1);
    ins(I_STAzp, ZPAddrInstLo);
    ins(I_CLC);
    ins(I_ADCimm, 0xFE);
    ins(I_TAX);
    ins(I_LDAzp, src2);
    ins(I_STAzp, ZPAddrInstHi);
    ins(I_ADCimm, 0xFF);
    ins(I_TAY);
    ins(I_LPA);
    ins(I_STAzp, ZPAddrClssLo);
    ins(I_LPA);
    ins(I_STAzp, ZPAddrClssHi);
    ins(I_PLX);*/
  }

  public void genCall(int off, int clssReg, int parSize) {
    /*int cr1, cr2;
    if ((cr1=getRegZPAddr(1, clssReg, StdTypes.T_PTR, false))==-1
        || (cr2=getRegZPAddr(2, clssReg, StdTypes.T_PTR, false))==-1) return;
    ins(I_PHX);
    ins(I_CLC);
    ins(I_LDAzp, cr1);
    ins(I_ADCimm, off&0xFF);
    ins(I_TAX);
    ins(I_LDAzp, cr2);
    ins(I_ADCimm, off>>>8);
    ins(I_TAY);
    if (ctx.codeStart==0) {
      ins(I_LPA);
      ins(I_STAzp, ZPAddrTmpLo);
      ins(I_LPA);
      ins(I_STAzp, ZPAddrTmpHi);
      ins(I_PLX);
      ins(I_JSRmem, ZPAddrTmpLo);
    }
    else {
      ins(I_CLC);
      ins(I_LPA);
      ins(I_ADCimm, ctx.codeStart&0xFF);
      ins(I_STAzp, ZPAddrTmpLo);
      ins(I_LPA);
      ins(I_ADCimm, (ctx.codeStart>>>8)&0xFF);
      ins(I_STAzp, ZPAddrTmpHi);
      ins(I_PLX);
      ins(I_JSRmem, ZPAddrTmpLo);
    }
    insCleanStackAfterCall(parSize);*/
  }

  public void genCallIndexed(int intfReg, int off, int parSize) {
    /*int cr1, cr2;
    if ((cr1=getRegZPAddr(3, intfReg, StdTypes.T_DPTR, false))==-1
        || (cr2=getRegZPAddr(4, intfReg, StdTypes.T_DPTR, false))==-1) return;
    ins(I_PHX);
    ins(I_CLC);
    ins(I_LDAzp, cr1);
    ins(I_ADCimm, off&0xFF);
    ins(I_TAX);
    ins(I_LDAzp, cr2);
    ins(I_ADCimm, off>>>8);
    ins(I_TAY);
    ins(I_LPA);
    ins(I_STAzp, ZPAddrTmpLo);
    ins(I_LPA);
    ins(I_STAzp, ZPAddrTmpHi);
    ins(I_CLC);
    ins(I_LDAzp, ZPAddrTmpLo);
    ins(I_ADCzp, ZPAddrClssLo);
    ins(I_TAX);
    ins(I_LDAzp, ZPAddrTmpHi);
    ins(I_ADCzp, ZPAddrClssHi);
    ins(I_TAY);
    if (ctx.codeStart==0) {
      ins(I_LPA);
      ins(I_STAzp, ZPAddrTmpLo);
      ins(I_LPA);
      ins(I_STAzp, ZPAddrTmpHi);
      ins(I_PLX);
      ins(I_JSRmem, ZPAddrTmpLo);
    }
    else {
      ins(I_CLC);
      ins(I_LPA);
      ins(I_ADCimm, ctx.codeStart&0xFF);
      ins(I_STAzp, ZPAddrTmpLo);
      ins(I_LPA);
      ins(I_ADCimm, (ctx.codeStart>>>8)&0xFF);
      ins(I_STAzp, ZPAddrTmpHi);
      ins(I_PLX);
      ins(I_JSRmem, ZPAddrTmpLo);
    }
    insCleanStackAfterCall(parSize);*/
  }

  public void genCallConst(Mthd obj, int parSize) {
    insPatchedCall(obj, parSize);
    insCleanStackAfterCall(parSize);
  }

  public void genJmp(Instruction dest) {
    //insJmp(I_JMPimm, dest);
  }

  public void genCondJmp(Instruction dest, int cond) {
    /*Instruction here;
    switch (cond) {
      case Ops.C_EQ: //"=="
        insJmp(I_JPZimm, dest);
        break;
      case Ops.C_NE: //"!="
        insJmp(I_JNZimm, dest);
        break;
      case Ops.C_LW: //"<"
      case Ops.C_BO: //unsigned "<" i.e. below (sign "not"-handled in genComp)
        insJmp(I_JNCimm, dest);
        break;
      case Ops.C_GE: //">="
        insJmp(I_JPCimm, dest);
        break;
      case Ops.C_LE: //"<="
        insJmp(I_JNCimm, dest);
        insJmp(I_JPZimm, dest);
        break;
      case Ops.C_GT: //">"
        here=getUnlinkedInstruction();
        insJmp(I_JPZimm, here);
        insJmp(I_JPCimm, dest);
        appendInstruction(here);
        break;
      default:
        fatalError("unsupported jump in genCondJmp");
    }*/
  }

  public int genComp(int src1, int src2, int type, int cond) {
    /*int i, max, r1, r2;
    Instruction zeroKeeper;
    max=getByteCount(type);
    if (cond!=Ops.C_EQ && cond!=Ops.C_NE && cond!=Ops.C_BO && type!=StdTypes.T_CHAR) {
      if ((r1=getRegZPAddr(max, src1, type, false))==-1 ||
          (r2=getRegZPAddr(max, src2, type, false))==-1) return 0;
      ins(I_LDAzp, r1);
      ins(I_EORimm, 0x80);
      ins(I_STAzp, ZPAddrTmpLo);
      ins(I_LDAzp, r2);
      ins(I_EORimm, 0x80);
      ins(I_STAzp, ZPAddrTmpHi);
    }
    if (max>1) ins(I_CLY);
    for (i=1; i<=max; i++) {
      if (i==max && cond!=Ops.C_EQ && cond!=Ops.C_NE && cond!=Ops.C_BO && type!=StdTypes.T_CHAR) {
        r1=ZPAddrTmpLo;
        r2=ZPAddrTmpHi;
      }
      else if ((r1=getRegZPAddr(i, src1, type, false))==-1 ||
          (r2=getRegZPAddr(i, src2, type, false))==-1) return 0;
      ins(I_LDAzp, r1);
      ins(i==1 ? I_CMPzp : I_SBCzp, r2);
      if (max>1) {
        zeroKeeper=getUnlinkedInstruction();
        insJmp(I_JPZimm, zeroKeeper);
        ins(I_INY);
        appendInstruction(zeroKeeper);
      }
    }
    if (max>1) ins(I_TYA);*/
    return cond;
  }

  public int genCompValToConstVal(int src, int val, int type, int cond) {
    /*int i, max, r;
    Instruction zeroKeeper;
    max=getByteCount(type);
    if (cond!=Ops.C_EQ && cond!=Ops.C_NE && cond!=Ops.C_BO) {
      if ((r=getRegZPAddr(max, src, type, false))==-1) return 0;
      ins(I_LDAzp, r);
      ins(I_EORimm, 0x80);
      ins(I_STAzp, ZPAddrTmpLo);
    }
    if (max>1) ins(I_CLY);
    for (i=1; i<=max; i++) {
      if ((r=getRegZPAddr(i, src, type, false))==-1) return 0;
      if (i==max && cond!=Ops.C_EQ && cond!=Ops.C_NE && cond!=Ops.C_BO) {
        ins(I_LDAzp, ZPAddrTmpLo);
        ins(i==1 ? I_CMPimm : I_SBCimm, ((val>>>((i-1)<<3))&0xFF)^0x80);
      }
      else {
        ins(I_LDAzp, r);
        ins(i==1 ? I_CMPimm : I_SBCimm, (val>>>((i-1)<<3))&0xFF);
      }
      if (max>1) {
        zeroKeeper=getUnlinkedInstruction();
        insJmp(I_JPZimm, zeroKeeper);
        ins(I_INY);
        appendInstruction(zeroKeeper);
      }
    }
    if (max>1) ins(I_TYA);*/
    return cond;
  }

  public int genCompValToConstDoubleOrLongVal(int src, long val, boolean asDouble, int cond) {
    /*int i, r;
    Instruction zeroKeeper;
    if (cond!=Ops.C_EQ && cond!=Ops.C_NE && cond!=Ops.C_BO) {
      if ((r=getRegZPAddr(8, src, StdTypes.T_LONG, false))==-1) return 0;
      ins(I_LDAzp, r);
      ins(I_EORimm, 0x80);
      ins(I_STAzp, ZPAddrTmpLo);
    }
    ins(I_CLY);
    for (i=1; i<=8; i++) {
      if ((r=getRegZPAddr(i, src, StdTypes.T_LONG, false))==-1) return 0;
      if (i==1 && cond!=Ops.C_EQ && cond!=Ops.C_NE && cond!=Ops.C_BO) {
        ins(I_LDAzp, ZPAddrTmpLo);
        ins(i==1 ? I_CMPimm : I_SBCimm, ((int)((val>>>((i-1)<<3)))&0xFF)^0x80);
      }
      else {
        ins(I_LDAzp, r);
        ins(i==1 ? I_CMPimm : I_SBCimm, ((int)(val>>>((i-1)<<3)))&0xFF);
      }
      zeroKeeper=getUnlinkedInstruction();
      insJmp(I_JPZimm, zeroKeeper);
      ins(I_INY);
      appendInstruction(zeroKeeper);
    }
    ins(I_TYA);*/
    return cond;
  }

  public int genCompPtrToNull(int reg, int cond) {
    /*int cr1, cr2;
    if ((cr1=getRegZPAddr(1, reg, StdTypes.T_PTR, false))==-1
        || (cr2=getRegZPAddr(2, reg, StdTypes.T_PTR, false))==-1) return 0;
    ins(I_LDAzp, cr1);
    ins(I_ORzp, cr2);*/
    return cond;
  }

  public void genWriteIO(int dst, int src, int type, int memLoc) {
    // TODO Auto-generated method stub
  }

  public void genReadIO(int dst, int src, int type, int memLoc) {
    // TODO Auto-generated method stub
  }

  public void genMoveToPrimary(int src, int type) {
    /*int tempReg, primReg, i=1, count;
    for (count=getByteCount(type); i<=count; i++) {
      if ((tempReg=getRegZPAddr(i, src, type, false))==-1) return;
      primReg=getZPAddrOfReg(i-1);
      if (tempReg!=primReg) ins(I_MOVzpzp, primReg, tempReg);
    }*/
  }

  public void genMoveFromPrimary(int dst, int type) {
    /*int tempReg, primReg, i=getByteCount(type);
    while (i>0) {
      if ((tempReg=getRegZPAddr(i, dst, type, true))==-1) return;
      writtenRegs|=(1<<tempReg);
      primReg=getZPAddrOfReg(--i);
      if (primReg!=tempReg) ins(I_MOVzpzp, tempReg, primReg);
    }*/
  }

  public void genMoveIntfMapFromPrimary(int dst) {
    /*int r1, r2;
    if ((r1=getRegZPAddr(3, dst, StdTypes.T_DPTR, true))==0
        || (r2=getRegZPAddr(4, dst, StdTypes.T_DPTR, true))==0) return;
    ins(I_MOVzpzp, r1, getZPAddrOfReg(0)); //interface map double pointer
    ins(I_MOVzpzp, r2, getZPAddrOfReg(1));*/ //  is never in r0/r1, so always move
  }

  public void genSavePrimary(int type) {
    /*int primReg, count;
    count=getByteCount(type);
    for (primReg=0; count>0; primReg++, count--) ins(I_PUSHzp, getZPAddrOfReg(primReg));*/
  }

  public void genRestPrimary(int type) {
    /*int primReg, count;
    count=getByteCount(type);
    for (primReg=count-1; count>0; primReg--, count--) ins(I_POPzp, getZPAddrOfReg(primReg));*/
  }

  public void genCheckBounds(int addrReg, int offReg, int checkToOffset, Instruction onSuccess) {
    /*int addrR1, addrR2, offR1, offR2; //although offReg is of type INT, we use only lower 2 bytes
    if ((addrR1=getRegZPAddr(1, addrReg, StdTypes.T_PTR, false))==-1
        || (addrR2=getRegZPAddr(2, addrReg, StdTypes.T_PTR, false))==-1
        || (offR1=getRegZPAddr(1, offReg, StdTypes.T_PTR, false))==-1
        || (offR2=getRegZPAddr(2, offReg, StdTypes.T_PTR, false))==-1) return;
    ins(I_PHX);
    ins(I_CLC);
    ins(I_LDAzp, addrR1);
    ins(I_ADCimm, checkToOffset&0xFF);
    ins(I_TAX);
    ins(I_LDAzp, addrR2);
    ins(I_ADCimm, (checkToOffset>>>8)&0xFF);
    ins(I_TAY);
    ins(I_LPA);
    ins(I_STAzp, ZPAddrTmpLo);
    ins(I_LPA);
    ins(I_STAzp, ZPAddrTmpHi);
    ins(I_PLX);
    ins(I_LDAzp, offR1);
    ins(I_CMPzp, ZPAddrTmpLo);
    ins(I_LDAzp, offR2);
    ins(I_SBCzp, ZPAddrTmpHi);
    insJmp(I_JNCimm, onSuccess);*/ //if (off<limit) jump everything is ok
  }

  public void genCheckStackExtreme(int maxValueReg, Instruction onSuccess) {
    fatalError("stack extreme check not supported");
  }

  public void genLoadDerefAddr(int destReg, int objReg, int indReg, int baseOffset, int entrySize) {
    /*int dstR1, dstR2, objR1, objR2, indR1, indR2;
    if ((dstR1=getRegZPAddr(1, destReg, StdTypes.T_SHRT, true))==-1
        || (dstR2=getRegZPAddr(2, destReg, StdTypes.T_SHRT, true))==-1
        || (objR1=getRegZPAddr(1, objReg, StdTypes.T_SHRT, false))==-1
        || (objR2=getRegZPAddr(2, objReg, StdTypes.T_SHRT, false))==-1
        || (indR1=getRegZPAddr(1, indReg, StdTypes.T_SHRT, false))==-1
        || (indR2=getRegZPAddr(2, indReg, StdTypes.T_SHRT, false))==-1) return;
    if (entrySize<0) {
      ins(I_CLA);
      ins(I_SEC);
      ins(I_SBCzp, indR1);
      ins(I_STAzp, ZPAddrTmpLo);
      ins(I_CLA);
      ins(I_SBCzp, indR2);
      ins(I_STAzp, ZPAddrTmpHi);
      entrySize=-entrySize;
    }
    else {
      ins(I_MOVzpzp, ZPAddrTmpLo, indR1);
      ins(I_MOVzpzp, ZPAddrTmpHi, indR2);
    }
    switch (entrySize) {
      case 1: case 2: case 4: case 8:
        while (entrySize>1) {
          ins(I_SHLzp, ZPAddrTmpLo);
          ins(I_ROLzp, ZPAddrTmpHi);
          entrySize=entrySize>>>1;
        }
        break;
      default:
        fatalError("not supported entrySize in genLoadDerefAddr");
        return;
    }
    if (baseOffset!=0) {
      ins(I_CLC);
      ins(I_LDAimm, baseOffset&0xFF);
      ins(I_ADCzp, ZPAddrTmpLo);
      ins(I_STAzp, ZPAddrTmpLo);
      ins(I_LDAimm, (baseOffset>>>8)&0xFF);
      ins(I_ADCzp, ZPAddrTmpHi);
      ins(I_STAzp, ZPAddrTmpHi);
    }
    ins(I_CLC);
    ins(I_LDAzp, objR1);
    ins(I_ADCzp, ZPAddrTmpLo);
    ins(I_STAzp, dstR1);
    ins(I_LDAzp, objR2);
    ins(I_ADCzp, ZPAddrTmpHi);
    ins(I_STAzp, dstR2);*/
  }
  
  //throw frame:
  //  excOffset + 10: current exception thrown
  //  excOffset +  9: base pointer for current try-block
  //  excOffset +  8: stack pointer for current try-block
  //  excOffset +  6: instance context of current try-block
  //  excOffset +  4: unit context of current try-block
  //  excOffset +  2: code-byte to jump to if exception is thrown
  //  excOffset     : pointer to last excStackFrame
  public void genThrowFrameBuild(int globalAddrReg, Instruction dest, int throwBlockOffset) {
    if (globalAddrReg!=0x0003 || usedRegs!=0x0003 || throwBlockOffset>=0) { //globalAddrReg is the only allocated register, throwBlockOffset is in local vars
      fatalError(ERR_INVGLOBADDRREG);
      return;
    }
    /*throwBlockOffset+=STACKMEM_OFF;
    //get global frame pointer, save in current frame, set current frame as global
    ins(I_CLY); //load current global pointer and store throwBlockOffset+0/1
    ins(I_LDAizpy, ZPAddrBase);
    ins(I_STAabsx, throwBlockOffset);
    ins(I_INY);
    ins(I_LDAizpy, ZPAddrBase);
    ins(I_STAabsx, throwBlockOffset+1);
    ins(I_TXA); //calculate current throw frame address, store in global pointer
    ins(I_CLY);
    ins(I_CLC);
    ins(I_ADCimm, throwBlockOffset&0xFF);
    ins(I_STAizpy, ZPAddrBase);
    ins(I_CLA);
    ins(I_ADCimm, (throwBlockOffset>>>8)&0xFF);
    ins(I_INY);
    ins(I_STAizpy, ZPAddrBase);
    //insert destination code address
    insJmp(I_PUSAimm, dest);
    ins(I_PLA);
    ins(I_STAabsx, throwBlockOffset+2);
    ins(I_PLA);
    ins(I_STAabsx, throwBlockOffset+3);
    //fill unit- and instance context, stack and base pointer
    ins(I_LDAzp, ZPAddrClssLo);
    ins(I_STAabsx, throwBlockOffset+4);
    ins(I_LDAzp, ZPAddrClssHi);
    ins(I_STAabsx, throwBlockOffset+5);
    ins(I_LDAzp, ZPAddrInstLo);
    ins(I_STAabsx, throwBlockOffset+6);
    ins(I_LDAzp, ZPAddrInstHi);
    ins(I_STAabsx, throwBlockOffset+7);
    ins(I_TXY);
    ins(I_TSX);
    ins(I_TXA);
    ins(I_TYX);
    ins(I_STAabsx, throwBlockOffset+8);
    ins(I_TXA);
    ins(I_STAabsx, throwBlockOffset+9);
    //clear exception thrown
    ins(I_CLA);
    ins(I_STAabsx, throwBlockOffset+10);
    ins(I_STAabsx, throwBlockOffset+11);*/
  }

  public void genThrowFrameUpdate(Instruction oldDest, Instruction newDest, int throwBlockOffset) {
    if (throwBlockOffset>=0) { //throwBlockOffset is in local vars
      fatalError(ERR_INVGLOBADDRREG);
      return;
    }
    /*throwBlockOffset+=STACKMEM_OFF;
    insJmp(I_PUSAimm, newDest);
    ins(I_PLA);
    ins(I_STAabsx, throwBlockOffset+2);
    ins(I_PLA);
    ins(I_STAabsx, throwBlockOffset+3);*/
  }

  public void genThrowFrameReset(int globalAddrReg, int throwBlockOffset) {
    if (globalAddrReg!=0x0003 || usedRegs!=0x0003 || throwBlockOffset>=0) { //globalAddrReg is the only allocated register, throwBlockOffset is in local vars
      fatalError(ERR_INVGLOBADDRREG);
      return;
    }
    //load previous pointer from current frame and store in global variable
    /*throwBlockOffset+=STACKMEM_OFF;
    ins(I_CLY);
    ins(I_LDAabsx, throwBlockOffset);
    ins(I_STAizpy, ZPAddrBase);
    ins(I_INY);
    ins(I_LDAabsx, throwBlockOffset+1);
    ins(I_STAizpy, ZPAddrBase);*/
  }

  public void inlineVarOffset(int inlineMode, int objReg, Object loc, int offset, int baseValue) {
    fatalError("inlining of variable offsets is not supported");
  }
  
  public void inlineCodeAddress(boolean defineHere, int addOffset) {
    /*Instruction i;
    appendInstruction(i=getUnlinkedInstruction());
    if (defineHere) i.type=IT_DEFICA;
    else {
      i.type=IT_INLCODA;
      i.iPar1=addOffset;
      i.putShort(0); //dummy address
    }*/
  }
  
  //optimized storing and calculation
  public void genStoreVarVal(int objReg, Object loc, int off, int valReg, int type) {
    /*int valR, objR, count, i;
    count=getByteCount(type);
    int pos=mem.getAddrAsInt(loc, off);
    if (objReg==regBase) {
      if (pos>=0) pos+=curVarOffParam;
      pos+=STACKMEM_OFF;
      for (i=0; i<count; i++) {
        if ((valR=getRegZPAddr(i+1, valReg, type, false))==-1) return;
        ins(I_LDAzp, valR);
        ins(I_STAabsx, pos+i);
      }
    }
    else {
      ins(I_PHX);
      if (objReg==0) {
        ins(I_LDXimm, pos&0xFF);
        ins(I_LDYimm, pos>>>8);
      }
      else {
        ins(I_CLC);
        if ((objR=getRegZPAddr(1, objReg, StdTypes.T_PTR, false))==-1) return;
        ins(I_LDAzp, objR);
        ins(I_ADCimm, pos&0xFF);
        ins(I_TAX);
        if ((objR=getRegZPAddr(2, objReg, StdTypes.T_PTR, false))==-1) return;
        ins(I_LDAzp, objR);
        ins(I_ADCimm, pos>>>8);
        ins(I_TAY);
      }
      for (i=0; i<count; i++) {
        if ((valR=getRegZPAddr(i+1, valReg, type, false))==-1) return;
        ins(I_LDAzp, valR);
        ins(I_SPA);
      }
      ins(I_PLX);
    }*/
  }
  
  public void genStoreVarConstVal(int objReg, Object loc, int off, int val, int type) {
    /*int objR, count, i;
    count=getByteCount(type);
    int pos=mem.getAddrAsInt(loc, off);
    if (objReg==regBase) {
      if (pos>=0) pos+=curVarOffParam;
      pos+=STACKMEM_OFF;
      for (i=0; i<count; i++) {
        ins(I_LDAimm, (val>>>(i<<3))&0xFF);
        ins(I_STAabsx, pos+i);
      }
    }
    else {
      ins(I_PHX);
      if (objReg==0) {
        ins(I_LDXimm, pos&0xFF);
        ins(I_LDYimm, pos>>>8);
      }
      else {
        ins(I_CLC);
        if ((objR=getRegZPAddr(1, objReg, StdTypes.T_PTR, false))==-1) return;
        ins(I_LDAzp, objR);
        ins(I_ADCimm, pos&0xFF);
        ins(I_TAX);
        if ((objR=getRegZPAddr(2, objReg, StdTypes.T_PTR, false))==-1) return;
        ins(I_LDAzp, objR);
        ins(I_ADCimm, pos>>>8);
        ins(I_TAY);
      }
      for (i=0; i<count; i++) {
        ins(I_LDAimm, (val>>>(i<<3))&0xFF);
        ins(I_SPA);
      }
      ins(I_PLX);
    }*/
  }
  
  public void genStoreVarConstDoubleOrLongVal(int objReg, Object loc, int off, long val, boolean asDouble) {
    /*int objR, i;
    int pos=mem.getAddrAsInt(loc, off);
    if (objReg==regBase) {
      if (pos>=0) pos+=curVarOffParam;
      pos+=STACKMEM_OFF;
      for (i=0; i<8; i++) {
        ins(I_LDAimm, ((int)(val>>>(i<<3)))&0xFF);
        ins(I_STAabsx, pos+i);
      }
    }
    else {
      ins(I_PHX);
      if (objReg==0) {
        ins(I_LDXimm, pos&0xFF);
        ins(I_LDYimm, pos>>>8);
      }
      else {
        ins(I_CLC);
        if ((objR=getRegZPAddr(1, objReg, StdTypes.T_PTR, false))==-1) return;
        ins(I_LDAzp, objR);
        ins(I_ADCimm, pos&0xFF);
        ins(I_TAX);
        if ((objR=getRegZPAddr(2, objReg, StdTypes.T_PTR, false))==-1) return;
        ins(I_LDAzp, objR);
        ins(I_ADCimm, pos>>>8);
        ins(I_TAY);
      }
      for (i=0; i<8; i++) {
        ins(I_LDAimm, ((int)(val>>>(i<<3)))&0xFF);
        ins(I_SPA);
      }
      ins(I_PLX);
    }*/
  }
  
  public void genBinOpConstRi(int dst, int src1, int val, int op, int type) { //may destroy src1
    /*int opType=op>>>16, opPar=op&0xFFFF;
    int dstR, srcR1, i, count, tmpVal;
    Instruction redo;
    count=getByteCount(type);
    switch (type) {
      case StdTypes.T_BOOL:
        if (opType!=Ops.S_ARI || !(opPar==Ops.A_AND || opPar==Ops.A_OR || opPar==Ops.A_XOR)) {
          fatalError("unsupported operation for bool-genBinOpConstRi");
          return;
        }
        //has to do the following, too!
      case StdTypes.T_BYTE:
        if (opPar==Ops.A_MUL) { //special case: mul for byte is supported
          if ((dstR=getRegZPAddr(1, dst, StdTypes.T_BYTE, true))==-1
              || (srcR1=getRegZPAddr(1, src1, StdTypes.T_BYTE, false))==-1) return;
          ins(I_PHX);
          ins(I_LDAzp, srcR1);
          ins(I_MULimm, val&0xFF);
          ins(I_STAzp, dstR);
          ins(I_PLX);
          return;
        }
        //has to do the following, too!
      case StdTypes.T_SHRT: case StdTypes.T_CHAR:
      case StdTypes.T_INT: case StdTypes.T_LONG:
        switch (opType) {
          case Ops.S_ARI:
            for (i=0; i<count; i++) {
              if ((dstR=getRegZPAddr(i+1, dst, type, true))==-1
                  || (srcR1=getRegZPAddr(i+1, src1, type, false))==-1) return;
              tmpVal=(val>>>(i<<3))&0xFF;
              ins(I_LDAzp, srcR1);
              switch (opPar) {
                case Ops.A_AND:
                  ins(I_ANDimm, tmpVal);
                  break;
                case Ops.A_OR:
                  ins(I_ORimm, tmpVal);
                  break;
                case Ops.A_XOR:
                  ins(I_EORimm, tmpVal);
                  break;
                case Ops.A_MINUS:
                  if (i==0) ins(I_SEC);
                  ins(I_SBCimm, tmpVal);
                  break;
                case Ops.A_PLUS:
                  if (i==0) ins(I_CLC);
                  ins(I_ADCimm, tmpVal);
                  break;
                default:
                  fatalError("unsupported ari-operation for genBinOpConstRi");
                  return;
              }
              ins(I_STAzp, dstR);
            }
            return;
          case Ops.S_BSH:
            if (val==0) return; //do not shift
            ins(I_LDYimm, val);
            redo=getUnlinkedInstruction();
            appendInstruction(redo);
            if ((dstR=getRegZPAddr(opPar==Ops.B_SHL ? 1 : count, dst, StdTypes.T_BYTE, true))==-1
                || (srcR1=getRegZPAddr(opPar==Ops.B_SHL ? 1 : count, src1, StdTypes.T_BYTE, false))==-1) return;
            switch (opPar) {
              case Ops.B_SHL:
                if (srcR1!=dstR) {
                  ins(I_LDAzp, srcR1);
                  ins(I_SHL);
                  ins(I_STAzp, dstR);
                }
                else ins(I_SHLzp, dstR);
                i=2;
                break;
              case Ops.B_SHRL:
                if (srcR1!=dstR) {
                  ins(I_LDAzp, srcR1);
                  ins(I_SHR);
                  ins(I_STAzp, dstR);
                }
                else ins(I_SHRzp, dstR);
                i=count-1;
                break;
              case Ops.B_SHRA:
                ins(I_LDAzp, srcR1);
                ins(I_SHL);
                ins(I_LDAzp, srcR1);
                ins(I_ROR);
                ins(I_STAzp, dstR);
                i=count-1;
                break;
              default:
                fatalError("unsupported bsh-operation for genBinOpConstRi");
                return;
            }
            while (--count>0) { //higher bytes - respect carry
              if ((dstR=getRegZPAddr(i, dst, type, true))==-1
                  || (srcR1=getRegZPAddr(i, src1, type, false))==-1) return;
              if (opPar==Ops.B_SHL) {
                if (dstR!=srcR1) {
                  ins(I_LDAzp, srcR1);
                  ins(I_ROL);
                  ins(I_STAzp, dstR);
                }
                else ins(I_ROLzp, dstR);
                i++;
              }
              else {
                if (dstR!=srcR1) {
                  ins(I_LDAzp, srcR1);
                  ins(I_ROR);
                  ins(I_STAzp, dstR);
                }
                else ins(I_RORzp, dstR);
                i--;
              }              
            }
            insJmp(I_DYJPimm, redo); //decrease counter and jump to redo
            return;
          default:
            fatalError("unsupported operation-type for genBinOpConstRi");
            return;
        }
      default:
        fatalError("unsupported operand-type for genBinOpConstRi");
    }*/
  }
  
  public void genBinOpConstDoubleOrLongRi(int dst, int src1, long val, int op, boolean asDouble) {
    /*int opType=op>>>16, opPar=op&0xFFFF;
    int dstR, srcR1, i, tmpVal, count;
    Instruction redo;
    if (asDouble) super.genBinOpConstDoubleOrLongRi(dst, src1, val, opPar, asDouble);
    switch (opType) {
      case Ops.S_ARI:
        for (i=0; i<8; i++) {
          if ((dstR=getRegZPAddr(i+1, dst, StdTypes.T_LONG, true))==-1
              || (srcR1=getRegZPAddr(i+1, src1, StdTypes.T_LONG, false))==-1) return;
          tmpVal=((int)(val>>>(i<<3)))&0xFF;
          ins(I_LDAzp, srcR1);
          switch (opPar) {
            case Ops.A_AND:
              ins(I_ANDimm, tmpVal);
              break;
            case Ops.A_OR:
              ins(I_ORimm, tmpVal);
              break;
            case Ops.A_XOR:
              ins(I_EORimm, tmpVal);
              break;
            case Ops.A_MINUS:
              if (i==0) ins(I_SEC);
              ins(I_SBCimm, tmpVal);
              break;
            case Ops.A_PLUS:
              if (i==0) ins(I_CLC);
              ins(I_ADCimm, tmpVal);
              break;
            default:
              fatalError("unsupported ari-operation for genBinOpConstDoubleOrLongRi");
              return;
          }
          ins(I_STAzp, dstR);
        }
        return;
      case Ops.S_BSH:
        if (val==0) return; //do not shift
        ins(I_LDYimm, (int)val&0xFF);
        redo=getUnlinkedInstruction();
        appendInstruction(redo);
        if ((dstR=getRegZPAddr(opPar==Ops.B_SHL ? 1 : 8, dst, StdTypes.T_LONG, true))==-1
            || (srcR1=getRegZPAddr(opPar==Ops.B_SHL ? 1 : 8, src1, StdTypes.T_LONG, false))==-1) return;
        switch (opPar) {
          case Ops.B_SHL:
            if (srcR1!=dstR) {
              ins(I_LDAzp, srcR1);
              ins(I_SHL);
              ins(I_STAzp, dstR);
            }
            else ins(I_SHLzp, dstR);
            i=2;
            break;
          case Ops.B_SHRL:
            if (srcR1!=dstR) {
              ins(I_LDAzp, srcR1);
              ins(I_SHR);
              ins(I_STAzp, dstR);
            }
            else ins(I_SHRzp, dstR);
            i=7;
            break;
          case Ops.B_SHRA:
            ins(I_LDAzp, srcR1);
            ins(I_SHL);
            ins(I_LDAzp, srcR1);
            ins(I_ROR);
            ins(I_STAzp, dstR);
            i=7;
            break;
          default:
            fatalError("unsupported bsh-operation for genBinOpConstDoubleOrLongRi");
            return;
        }
        count=8;
        while (--count>0) { //higher bytes - respect carry
          if ((dstR=getRegZPAddr(i, dst, StdTypes.T_LONG, true))==-1
              || (srcR1=getRegZPAddr(i, src1, StdTypes.T_LONG, false))==-1) return;
          if (opPar==Ops.B_SHL) {
            if (dstR!=srcR1) {
              ins(I_LDAzp, srcR1);
              ins(I_ROL);
              ins(I_STAzp, dstR);
            }
            else ins(I_ROLzp, dstR);
            i++;
          }
          else {
            if (dstR!=srcR1) {
              ins(I_LDAzp, srcR1);
              ins(I_ROR);
              ins(I_STAzp, dstR);
            }
            else ins(I_RORzp, dstR);
            i--;
          }              
        }
        insJmp(I_DYJPimm, redo); //decrease counter and jump to redo
        return;
      default:
        fatalError("unsupported operation-type for genBinOpConstDoubleOrLongRi");
        return;
    }*/
  }
  
  //internal instruction coding
  private void ins(int op) {
    Instruction i;
    appendInstruction(i=getUnlinkedInstruction());
    i.type=op;
    i.putByte(op);
    i.size=(op&SIZE_MASK)>>>SIZE_SHIFT;
  }
  
  private void ins(int op, int par) {
    Instruction i;
    appendInstruction(i=getUnlinkedInstruction());
    i.type=op;
    switch (op&PAR_P1M) {
      case PAR_IMM1:
        i.iPar1=par;
        break;
      case PAR_REG1:
        i.reg0=par;
        break;
      default:
        fatalError("invalid call to ins(op,par)");
    }
    i.putByte(op);
    i.size=(op&SIZE_MASK)>>>SIZE_SHIFT;
  }
  
  private void ins(int op, int par1, int par2) {
    Instruction i;
    appendInstruction(i=getUnlinkedInstruction());
    i.type=op;
    switch (op&PAR_P1M) {
      case PAR_IMM1:
        i.iPar1=par1;
        break;
      case PAR_REG1:
        i.reg0=par1;
        break;
      default:
        fatalError("invalid call to ins(op,par)");
    }
    switch (op&PAR_P2M) {
      case PAR_IMM2:
        i.iPar2=par2;
        break;
      case PAR_REG2:
        i.reg1=par2;
        break;
      default:
        fatalError("invalid call to ins(op,par)");
    }
    i.putByte(op);
    i.size=(op&SIZE_MASK)>>>SIZE_SHIFT;
  }
  
  private Instruction insPatchedCall(Mthd refMthd, int parSize) {
    Instruction i;
    appendInstruction(i=getUnlinkedInstruction());
    i.type=IT_CALL;
    //TODO
    i.refMthd=refMthd;
    //i.putByte(I_JSRimm);
    addToCodeRefFixupList(i, 1);
    return i;
  }
  
  private Instruction insJmp(int op, Instruction dest) {
    Instruction i;
    appendInstruction(i=getUnlinkedInstruction());
    i.type=op;
    i.putByte(op);
    i.size=(op&SIZE_MASK)>>>SIZE_SHIFT;
    i.jDest=dest;
    return i;
  }
  
  private Instruction insJmp(int op, int reg, Instruction dest) {
    Instruction i;
    appendInstruction(i=getUnlinkedInstruction());
    i.type=op;
    i.jDest=dest;
    i.reg0=reg;
    i.putByte(op);
    i.size=(op&SIZE_MASK)>>>SIZE_SHIFT;
    return i;
  }
  
  private void insCleanStackAfterCall(int parSize) {
    if (parSize!=0) { //clean parameters from stack
      //ins(I_INS, parSize);
    }
  }
  
  protected void printCode(TextPrinter v, Instruction first, String comment, boolean asmMode) {
    Instruction now;
    int insCnt=0;
    if (asmMode) ctx.printUniqueMethodName(v, mthdContainer);
    else {
      mthdContainer.owner.printNameWithOuter(v);
      v.print('.');
      mthdContainer.printNamePar(v);
    }
    v.print(':');
    if (comment!=null) {
      v.print(" //");
      v.println(comment);
    }
    else v.println();
    now=first;
    while (now!=null) {
      if (now.token!=null) {
        if (asmMode) v.print("    ");
        ctx.printSourceHint(v, now.token, ";");
      }
      if (asmMode && now.isDest) {
        printJumpDest(v, now);
        v.println(':');
      }
      if (now.type!=I_NONE) {
        insCnt+=print(v, now, asmMode);
        v.println();
      }
      now=now.next;
    }
    if (!asmMode) {
      v.print("//instruction count");
      if (comment!=null) {
        v.print(' ');
        v.print(comment);
      }
      v.print(": ");
      v.println(insCnt);
    }
  }
  
  private void printJumpDest(TextPrinter v, Instruction dest) {
    ctx.printUniqueMethodName(v, mthdContainer);
    v.print("$i");
    v.print(dest.instrNr);
  }
  
  private int print(TextPrinter v, Instruction i, boolean asmMode) {
    if (i.type==I_NONE) return 0;
    if (i.type==IT_CALL) return 0; //TODO
    v.print("  ");
    switch (i.type) {
      case I_RET:    v.print("RET"); break;
      case I_JPNZra: v.print("JPNZ"); break;
      case I_DECri:  v.print("DEC"); break;
      case I_INCDri: v.print("INCD"); break; 
      case I_PUSHrr: v.print("PUSH"); break;
      case I_POPrr:  v.print("POP"); break;
      case I_XORrr:  v.print("XOR"); break;
      case I_ADDDrr: v.print("ADDD"); break;
      case I_LDri:   v.print("LD"); break;
      default:
        v.print("---unknown instruction");
        break;
    }
    if ((i.type&PAR_P1M)!=0) {
      v.print(' ');
      switch (i.type&PAR_P1M) {
        case PAR_REG1: v.print('r'); v.print(i.reg0); break;
        case PAR_IMM1: v.print(i.iPar1); break;
        case PAR_ADR1: printJumpDest(v, i.jDest); break;
      }
      if ((i.type&PAR_P2M)!=0) {
        v.print(',');
        switch (i.type&PAR_P2M) {
          case PAR_REG2: v.print('r'); v.print(i.reg1); break;
          case PAR_IMM2: v.print(i.iPar2); break;
          case PAR_ADR2: printJumpDest(v, i.jDest); break;
        }
      }
    }
    return 1;
  }
}
